/*															*/
/* AtMega328p/@16MHz USART ECHO TEST						*/
/*															*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>

#define FIFO_SIZE 16				/* FIFO buffer size		*/
#define BAUD 9600					/* USART BAUD			*/

/* USART PORTS define	*/
#define UBRR  UBRR0
#define UCSRB UCSR0B
#define UDR   UDR0
#define RXEN  RXEN0
#define TXEN  TXEN0
#define RXCIE RXCIE0
#define UDRIE UDRIE0

/* USART interrupt control define	*/
#define USART_RECV_ENABLE()  ( UCSRB |=  _BV(RXCIE) )
#define USART_RECV_DISABLE() ( UCSRB &= ~_BV(RXCIE) )

#define USART_SEND_ENABLE()  ( UCSRB |=  _BV(UDRIE) )
#define USART_SEND_DISABLE() ( UCSRB &= ~_BV(UDRIE) )

typedef struct FIFO_t {				/* FIFO buffer struct	*/
	uint8_t idx_w;
	uint8_t idx_r;
	uint8_t count;
	uint8_t buff[FIFO_SIZE];
} FIFO;

static volatile FIFO txfifo;		/* send FIFO buffer 	*/
static volatile FIFO rxfifo;		/* recv FIFO buffer 	*/

/* UART RECV interrupt					*/
ISR( USART_RX_vect )
{
	uint8_t data = UDR;
	if ( rxfifo.count < FIFO_SIZE ) {
		rxfifo.count++;
		rxfifo.buff[ rxfifo.idx_w++ ] = data;
		if ( rxfifo.idx_w >= FIFO_SIZE ) rxfifo.idx_w = 0;
	}
}

/* UART SEND registr empty interrupt	*/
ISR( USART_UDRE_vect )
{
	UDR = txfifo.buff[ txfifo.idx_r++ ];
	if ( txfifo.idx_r >= FIFO_SIZE ) txfifo.idx_r = 0;
	if ( --txfifo.count == 0 ) USART_SEND_DISABLE();
}

void usart_init( void )
{
	memset( (void *)&txfifo ,0x00 ,FIFO_SIZE );	/* TX FIFO buffer clear		*/
	memset( (void *)&rxfifo ,0x00 ,FIFO_SIZE );	/* RX FIFO buffer clear		*/
	UBRR = (uint16_t)( F_CPU / ( 16UL * (uint32_t)BAUD ) ) - 1;	/* set BAUD */
	UCSRB = _BV(RXEN) | _BV(TXEN);			/* RX / TX ENABLE				*/
	USART_RECV_ENABLE();					/* USART recv interrupt enable	*/
}

uint8_t usart_recv_test( void )
{
	return ( rxfifo.count );
}

uint8_t usart_getchar( void )
{
	uint8_t data;
	while ( !usart_recv_test() );
	USART_RECV_DISABLE();
	data = rxfifo.buff[ rxfifo.idx_r++ ];
	if ( rxfifo.idx_r >= FIFO_SIZE ) rxfifo.idx_r = 0;
	rxfifo.count--;
	USART_RECV_ENABLE();
	return ( data );
}

uint8_t usart_send_test( void )
{
	return ( FIFO_SIZE - txfifo.count );	/* calc free size */
}

void usart_putchar( uint8_t data )
{
	while ( !usart_send_test() );
	USART_SEND_DISABLE();
	txfifo.count++;
	txfifo.buff[ txfifo.idx_w++ ] = data;
	if ( txfifo.idx_w >= FIFO_SIZE ) txfifo.idx_w = 0;
	USART_SEND_ENABLE();
}

static void hardwareInit( void )
{
	/* PortB  +--------7: Pullup : non use  */
	/*        |+-------6: Pullup : non use  */
	/*        ||+------5: Pullup : non use  */
	/*        |||+-----4: Pullup : non use  */
	/*        ||||+----3: Pullup : non use  */
	/*        |||||+---2: Pullup : non use  */
	/*        ||||||+--1: Pullup : non use  */
	/*        |||||||+-0: Pullup : non use	*/
	DDRB  = 0b00000000;
	PORTB = 0b11111111;

	/* PortC  +--------7: Pullup : non use 	*/
	/*        |+-------6: Pullup : non use 	*/
	/*        ||+------5: Pullup : non use 	*/
	/*        |||+-----4: Pullup : non use 	*/
	/*        ||||+----3: Pullup : non use 	*/
	/*        |||||+---2: Pullup : non use 	*/
	/*        ||||||+--1: Pullup : non use	*/
	/*        |||||||+-0: Pullup : non use	*/
	DDRC  = 0b00000000;
	PORTC = 0b11111111;

	/* PortD  +--------7: Pullup : non use	*/
	/*        |+-------6: Pullup : non use	*/
	/*        ||+------5: Pullup : SW1		*/
	/*        |||+-----4: Pullup : SW2 		*/
	/*        ||||+----3: Pullup : non use	*/
	/*        |||||+---2: Pullup : non use  */
	/*        ||||||+--1: Out(L) : TXD      */
	/*        |||||||+-0: Pullup : RXD      */
	DDRD  = 0b00000010;
	PORTD = 0b11111101;

	SMCR = _BV(SE); /* Power seve enabel for IDOL */

	usart_init(); /* initialize USART		*/
	sei();		  /*enable global interrupt */
}

int main( void )
{
	hardwareInit();

	while ( 1 ) {
		asm( "sleep" );									/* CPU IDOL MODE	*/

		if ( usart_recv_test() != 0 ) {					/* ECHO BACK		*/
			uint8_t data = usart_getchar();
			usart_putchar( data );
			if ( data == '\r' ) usart_putchar( '\n' );	/* CR -> CR+LF		*/
		}
	}

	return ( 0 );
}
